#include "SevSeg.h"
#include <math.h>
#include <avr/sleep.h>

// Steps
#define CAPACITY        1
#define COUNT           2
#define TEETH           3
#define END             4
#define PARTY           5
#define BLANK           6

// Timings
#define COUNT_SEC       20            // Countdown duration (seconds)
#define BREAK_SEC       2             // Break display duration (seconds)
#define PARTY_SEC       4             // Party display duration (seconds)

// Buzzer settings
#define PIEZO_PIN       12            // Piezo speaker/buzzer pin
#define TEMPO           200           // BPM

// Voltage sensor
#define VOLTAGE_PIN     A0
#define SAMPLES         8
#define CELLS_IN_SERIES 1

// Note values
#define WH              60000 / TEMPO // Whole note
#define DB              2 * WH        // Double note
#define HA              WH / 2        // Half note
#define QR              WH / 4        // Quarter note
#define EG              WH / 8        // Eighth note
#define SX              WH / 16       // Sixteenth note

// Note frequencies
#define C5              523
#define E5              659
#define G5              784
#define C6              1047
#define G6              1568
#define D7              2349

// Definitions
#define SECOND          1000          // 1s = 1000ms
#define SPLIT_SECOND    125

SevSeg _sevseg;
bool _beep = 1;
float _capacity = 0.0;
static int _tooth;
static unsigned long _offset;

//  AAA
// F   B
// F   B
//  GGG
// E   C
// E   C
//  DDD

static const byte BAR_MAP[] = {
 //.GFEDCBA
  B00000000,
  B00001000,
  B01001000,
  B01001001
};

static const byte TOOTH_MAP[] = {
 //.GFEDCBA
  B01010101,
  B01101010,
  B00001001,
  B00001001
};

void setup(void) {
  byte numDigits = 2;
  // Digits:
  byte digitPins[] = {2, 3};
  // Segments:          A  B  C  D  E  F  G   Period
  byte segmentPins[] = {4, 5, 6, 7, 8, 9, 10, 11 };
  bool resistorsOnSegments = false;
  byte hardwareConfig = COMMON_ANODE;
  bool updateWithDelays = false;
  bool leadingZeros = false;
  bool disableDecPoint = true;
  
  _sevseg.begin(hardwareConfig, numDigits, digitPins, segmentPins, resistorsOnSegments, updateWithDelays, leadingZeros, disableDecPoint);
  _sevseg.setBrightness(100);

  _capacity = getBatCapacity(CELLS_IN_SERIES, getVoltage(VOLTAGE_PIN, SAMPLES));
}

void loop(void) {
  int counter;
  int bars;
  byte digitCodes[2];
  switch(getStep()) {
    case CAPACITY:
      bars = _capacity * 6;
      digitCodes[0] = BAR_MAP[max(bars - 3, 0)];
      digitCodes[1] = BAR_MAP[min(bars, 3)];
      _sevseg.setSegments(digitCodes);
      break;
    case COUNT:
      counter = (millis() - _offset) / SECOND;
      _sevseg.setNumber(COUNT_SEC - counter, 0);
      break;
    case TEETH:
      if (_beep) {
        upTone();
        _beep = false;
      }
      digitCodes[0] = TOOTH_MAP[3 - _tooth % 4];
      digitCodes[1] = TOOTH_MAP[_tooth % 4];
      _sevseg.setSegments(digitCodes);
      break;
    case END:
      if (_beep) {
        upTone();
        _beep = false;
      }
      _sevseg.setChars("HH");
      break;
    case PARTY:
      if (_beep) {
        partyTone();
        _beep = false;
      }
      counter = millis() / SPLIT_SECOND;
      digitCodes[0] = TOOTH_MAP[3 - counter % 4];
      digitCodes[1] = TOOTH_MAP[counter % 4];
      _sevseg.setSegments(digitCodes);
      break;
    case BLANK:
      goToSleep();
      break;
  }
  _sevseg.refreshDisplay(); 
}

unsigned int getStep(void) {
  int sec = millis() / SECOND;
  if (sec < BREAK_SEC) {
    return CAPACITY;
  }
  if (sec < 2 * BREAK_SEC) {
    _tooth = 0;
    _offset = millis();
    return TEETH;
  }
  if (sec < 2 * BREAK_SEC + COUNT_SEC) {
    _beep = 1;
    return COUNT;
  }
  if (sec < 3 * BREAK_SEC + COUNT_SEC)
  {
    _tooth = 1;
    _offset = millis();
    return TEETH;
  }
  if (sec < 3 * BREAK_SEC + 2 * COUNT_SEC) {
    _beep = 1;
    return COUNT;
  }
  if (sec < 4 * BREAK_SEC + 2 * COUNT_SEC)
  {
    _tooth = 2;
    _offset = millis();
    return TEETH;
  }
  if (sec < 4 * BREAK_SEC + 3 * COUNT_SEC) {
    _beep = 1;
    return COUNT;
  }
  if (sec < 5 * BREAK_SEC + 3 * COUNT_SEC)
  {
    _tooth = 3;
    _offset = millis();
    return TEETH;
  }
  if (sec < 5 * BREAK_SEC + 4 * COUNT_SEC) {
    _beep = 1;
    return COUNT;
  }
  if (sec < 6 * BREAK_SEC + 4 * COUNT_SEC) {
    _offset = millis();
    return END;
  }
  if (sec < 6 * BREAK_SEC + 5 * COUNT_SEC) {
    _beep = 1;
    return COUNT;
  }
  if (sec < 6 * BREAK_SEC + 5 * COUNT_SEC + PARTY_SEC) {
    return PARTY;
  }
  return BLANK;
}

void partyTone(void) {
  int upTone[14] = {C5, HA, E5, HA, G5, HA, C6, HA, 0, HA, G5, HA, C6, DB};
  play(upTone, 14);
}

void upTone(void) {
  int upTone[4] = {G6, QR, D7, EG};
  play(upTone, 4);
}

void play(int freq[], int siz) {
  _sevseg.blank();
  for ( int i = 0 ; i < siz ; i=i+2 ) {
    if (freq[i] == 0) {
      delay(freq[i + 1]);
    } else {
      pwmWave(PIEZO_PIN, freq[i], freq[i + 1]);
    }
  }
}

void pwmWave(int pin, int frequency, int duration) {
  unsigned long t = millis();
  bool flag = true;
  int us = round(1000000 / (2 * frequency));
  while (millis() - t < duration) {
    analogWrite(pin, flag ? 255 : 0);
    flag = !flag;
    delayMicroseconds(us);
  }
  analogWrite(pin, 0);
}

float getVoltage(int pin, int samples) {
  int val = 0;
  for (int i = 0; i < samples; i++) {
    val += analogRead(pin);
    delay(50);
  }
  return val / samples * (5.0 / 1023.0);
}

float getBatCapacity(int s, float v) {
  float capa = 1.25 * v / s - 3.875;
  return capa < 0 ? 0 : (capa > 1 ? 1 : capa);
}

void goToSleep(void) {
  _sevseg.blank();
  _sevseg.refreshDisplay(); 
  sleep_enable();
  set_sleep_mode (SLEEP_MODE_PWR_DOWN);
  digitalWrite(LED_BUILTIN, LOW);
  delay(SECOND);
  sleep_cpu();
}
